math.clamp = function(val, min, max)
    return math.min(math.max(val, min), max)
end

function math.fract(n)
    local x = math.floor(n)
    return n - x
end

function math.lerp(a, b, t)
    return a * (1 - t) + b * t
end

function math.sawtooth(t, T)
	return (T * t) % 1.0
end

function math.smoothstep(edge0, edge1, x)
	x = math.clamp((x - edge0) / (edge1 - edge0), 0, 1)
	return x * x * (3 - 2 * x)
end

local max_samples = 10
function math.bezier(p1x, p1y, p2x, p2y)
	local start_gradient_, end_gradient_ = 0, 0
	if (p1x > 0) then
		start_gradient_ = p1y / p1x;
	elseif (p1y == 0 and p2x > 0) then
		start_gradient_ = p2y / p2x;
	elseif (p1y == 0 and p2y == 0) then
		start_gradient_ = 1;
	else
		start_gradient_ = 0;
	end
	if (p2x < 1) then
		end_gradient_ = (p2y - 1) / (p2x - 1);
	elseif (p2y == 1 and p1x < 1) then
		end_gradient_ = (p1y - 1) / (p1x - 1);
	elseif (p2y == 1 and p1y == 1) then
		end_gradient_ = 1;
	else
		end_gradient_ = 0;
	end
	local cx_ = 3 * p1x
	local bx_ = 3 * (p2x - p1x) - cx_
	local ax_ = 1 - cx_ - bx_
	local cy_ = 3 * p1y
	local by_ = 3 * (p2y - p1y) - cy_
	local ay_ = 1 - cy_ - by_

	local function SampleCurveX(t)
		return ((ax_ * t + bx_) * t + cx_) * t
	end

	local function SampleCurveY(t)
		return ((ay_ * t + by_) * t + cy_) * t
	end

	local function SampleCurveDerivativeX(t)
		return (3.0 * ax_ * t + 2.0 * bx_) * t + cx_
	end

	local delta_t = 1.0 / max_samples
	local spline_samples_ = {}
	for i = 1, max_samples do
		spline_samples_[i] = SampleCurveX(i * delta_t)
	end

	local function SolveCurveX(x)
		if x < 0.0 then
			return 0.0 + start_gradient_ * x;
		end
		if x > 1.0 then
			return 1.0 + end_gradient_ * (x - 1.0);
		end

		local t0, t1, t2, x2, d2, i = 0, 0, x, 0, 0, 0
		for i = 2, max_samples do
			if x <= spline_samples_[i] then
				t1 = delta_t * i
				t0 = t1 - delta_t
				t2 = t0 + (t1 - t0) * (x - spline_samples_[i-1]) / (spline_samples_[i] - spline_samples_[i-1])
				break
			end
		end
		for _ = 1, 4 do
			x2 = SampleCurveX(t2) - x
			if math.abs(x2) < 0.01 then
				return t2
			end
			d2 = SampleCurveDerivativeX(t2)
			if math.abs(d2) < 0.01 then
				break
			end
			t2 = t2 - x2 / d2
		end
		if math.abs(x2) < 0.01 then
			return t2
		end
		while t0 < t1 do
			x2 = SampleCurveX(t2)
			if math.abs(x2 - x) < 0.01 then
				return t2
			end
			if x > x2 then
				t0 = t2
			else
				t1 = t2
			end
			t2 = (t1 + t0) * 0.5
		end
		return t2
	end

	return function(t)
		return SampleCurveY(SolveCurveX(t))
	end
end

local fast_out_slow_in = math.bezier(0.4, 0.0, 0.2, 1.0)
function math.interval(t0, t1, t, tween)
	return t < t0 and 0 or t > t1 and 1 or (tween or fast_out_slow_in)((t - t0) / (t1 - t0))
end

function math.round(n)
    return math.floor(n + 0.5)
end

function math.slow_random(i)
    local t = { }
    for j = 1, i do
        t[' ' .. j] = j
    end
    for _, j in pairs(t) do
        return j
    end
end









-- Perlin noise
local hash = {
	208,34,231,213,32,248,233,56,161,78,24,140,71,48,140,254,245,255,
	247,247,40, 185,248,251,245,28,124,204,204,76,36,1,107,28,234,163,
	202,224,245,128,167,204,9,92,217,54,239,174,173,102,193,189,190,121,
	100,108,167,44,43,77,180,204,8,81,70,223,11,38,24,254,210,210,177,
	32,81,195,243,125,8,169,112,32,97,53,195,13,203,9,47,104,125,117,
	114,124,165,203,181,235,193,206,70,180,174,0,167,181,41,164,30,116,
	127,198,245,146,87,224,149,206,57,4,192,210,65,210,129,240,178,105,
	228,108,245,148,140,40,35,195,38,58,65,207,215,253,65,85,208,76,62,
	3,237,55,89,232,50,217,64,244,157,199,121,252,90,17,212,203,149,152,
	140,187,234,177,73,174,193,100,192,143,97,53,145,135,19,103,13,90,
	135,151,199,91,239,247,33,39,145,101,120,99,3,186,86,99,41,237,203,
	111,79,220,135,158,42,30,154,120,67,87,167,135,176,183,191,253,115,
	184,21,233,58,129,233,142,39,128,211,118,137,139,255,114,20,218,113,
	154,27,127,246,250,1,8,198,250,209,92,222,173,21,88,102,219
};
local function noise2(x, y, seed)
	local yindex = (y + seed) % 256;
	if (yindex < 0) then
		yindex += 256;
    end
	local xindex = (hash[yindex+1] + x) % 256;
	if (xindex < 0) then
		xindex += 256;
    end
	return hash[xindex+1];
end

local function lin_inter(x, y, s)
	return x + s * (y-x);
end

local function smooth_inter(x, y, s)
	return lin_inter(x, y, s * s * (3-2*s));
end

local function noise2d(x, y, seed)
	local x_int = math.floor(x);
	local y_int = math.floor(y);
	local x_frac = x - x_int;
	local y_frac = y - y_int;
	local s = noise2(x_int, y_int, seed);
	local t = noise2(x_int+1, y_int, seed);
	local u = noise2(x_int, y_int+1, seed);
	local v = noise2(x_int+1, y_int+1, seed);
	local low = smooth_inter(s, t, x_frac);
	local high = smooth_inter(u, v, x_frac);
	return smooth_inter(low, high, y_frac);
end

local function perlin2d(x, y, freq, depth, seed)
	local xa = x*freq;
	local ya = y*freq;
	local amp = 1.0;
	local fin = 0;
	local div = 0.0;

	for _ = 0, depth-1 do
		div += 256 * amp;
		fin += noise2d(xa, ya, seed) * amp;
		amp /= 2;
		xa *= 2;
		ya *= 2;
    end

	return fin/div;
end

function math.noise(x, y, f, d, s)
    return perlin2d(x, y, f or 0.1, d or 100, s or 0);
end

function math.derivative(f, x, h)
	h = h or 1e-5
	return (f(x + h) - f(x - h)) / (2 * h)
end

function math.eq(a, b, epsilon)
	return math.abs(a - b) < (epsilon or 1e-5)
end