---@class Color : Class
local Color = require("util.common.class")("Color")

function Color.HSL2RGB(h, s, l)
    if s == 0 then return l, l, l end
    local function to(p, q, t)
        if t < 0 then t = t + 1 end
        if t > 1 then t = t - 1 end
        if t < .16667 then return p + (q - p) * 6 * t end
        if t < .5 then return q end
        if t < .66667 then return p + (q - p) * (.66667 - t) * 6 end
        return p
    end
    local q = l < .5 and l * (1 + s) or l + s - l * s
    local p = 2 * l - q
    return to(p, q, h + .33334), to(p, q, h), to(p, q, h - .33334)
end

function Color.RGB2HSL(r, g, b)
    local max, min = math.max(r, g, b), math.min(r, g, b)
    local b = max + min
    local h = b / 2
    if max == min then return 0, 0, h end
    local s, l = h, h
    local d = max - min
    s = l > .5 and d / (2 - b) or d / b
    if max == r then h = (g - b) / d + (g < b and 6 or 0)
    elseif max == g then h = (b - r) / d + 2
    elseif max == b then h = (r - g) / d + 4
    end
    return h * .16667, s, l
end

function Color.RGB2HEX(r, g, b)
    return ((math.floor(r * 255) << 16) | (math.floor(g * 255) << 8) | (math.floor(b * 255)))
end

function Color.HSL2HEX(h, s, l)
    return Color.RGB2HEX(Color.HSL2RGB(h, s, l))
end

function Color.HEX2RGB(hex)
    return ((hex >> 16) & 0xff) / 255, ((hex >> 8) & 0xff) / 255, (hex & 0xff) / 255
end

function Color.HEX2HSL(hex)
    return Color.RGB2HSL(Color.HEX2RGB(hex))
end

function Color.CMYK2RGB(c, m, y, k)
    return (1 - math.min(1, c * (1 - k) + k)), (1 - math.min(1, m * (1 - k) + k)), (1 - math.min(1, y * (1 - k) + k))
end

function Color.Mix(c1, c2, t)
    return Color(c1.r * (1 - t) + c2.r * t, c1.g * (1 - t) + c2.g * t, c1.b * (1 - t) + c2.b * t, c1.a * (1 - t) + c2.a * t)
end

function Color:__init(r, g, b, a)
    if type(r) == "string" then
        if r:sub(1,1) == "#" then
            if #r == 4 then
                local hex = tonumber(r:sub(2), 16)
                self.r = ((hex >> 8) & 0xf) / 15
                self.g = ((hex >> 4) & 0xf) / 15
                self.b = (hex & 0xf) / 15
                self.a = 1
            elseif #r == 5 then
                local hex = tonumber(r:sub(2), 16)
                self.r = ((hex >> 12) & 0xf) / 15
                self.g = ((hex >> 8) & 0xf) / 15
                self.b = ((hex >> 4) & 0xf) / 15
                self.a = (hex & 0xf) / 15
            elseif #r == 7 then
                local hex = tonumber(r:sub(2), 16)
                self.r = ((hex >> 16) & 0xff) / 255
                self.g = ((hex >> 8) & 0xff) / 255
                self.b = (hex & 0xff) / 255
                self.a = 1
            elseif #r == 9 then
                local hex = tonumber(r:sub(2), 16)
                self.r = ((hex >> 24) & 0xff) / 255
                self.g = ((hex >> 16) & 0xff) / 255
                self.b = ((hex >> 8) & 0xff) / 255
                self.a = (hex & 0xff) / 255
            else
                error("Invalid color format: " .. r)
            end
        elseif r:sub(1,1) == "h" then
            local h, s, l = r:match("hsl%((%d+)[^%d]+(%d+)[^%d]+(%d+)[^%d]?%)")
            self.r, self.g, self.b = Color.HSL2RGB(h / 360, s / 100, l / 100)
            self.a = 1
        elseif r:sub(1,1) == "r" then
            local r, g, b = r:match("rgb%((%d+)[^%d]+(%d+)[^%d]+(%d+)[^%d]?%)")
            self.r = r / 255
            self.g = g / 255
            self.b = b / 255
            self.a = 1
        elseif r:sub(1,1) == "c" then
            local c, m, y, k = r:match("cmyk%((%d+)[^%d]+(%d+)[^%d]+(%d+)[^%d]+(%d+)[^%d]?%)")
            self.r, self.g, self.b = Color.CMYK2RGB(c / 100, m / 100, y / 100, k / 100)
            self.a = 1
        end
        return
    elseif type(r) == "table" then
        self.r = r.r or 0
        self.g = r.g or 0
        self.b = r.b or 0
        self.a = r.a or 1
        return
    end
    if r and not g and not b and not a then
        r = bit.tobit(r) & 0xffffffff
        self.r, self.g, self.b = Color.HEX2RGB(r)
        self.a = (r >> 24) & 0xff
        if self.a > 0 then
            self.a = self.a / 255
        else
            self.a = 1
        end
        return
    end
    self.r = r or 0
    self.g = g or 0
    self.b = b or 0
    self.a = a or 1
end

function Color:__tostring()
    return string.format("Color(%.2f, %.2f, %.2f, %.2f)", self.r, self.g, self.b, self.a)
end

function Color:__eq(other)
    return self.r == other.r and self.g == other.g and self.b == other.b and self.a == other.a
end

function Color:__add(other)
    return Color(self.r, self.g, self.b, self.a + other.a)
end

function Color:__sub(other)
    return Color(self.r, self.g, self.b, self.a - other.a)
end

function Color:__mul(other)
    return Color(self.r, self.g, self.b, self.a * other.a)
end

function Color:__div(other)
    return Color(self.r, self.g, self.b, self.a / other)
end

function Color:get(a)
    return Color.RGB2HEX(self.r, self.g, self.b) | (math.ceil((a or self.a) * 255) << 24)
end

function Color.__properties.color:get()
    return Color.RGB2HEX(self.r, self.g, self.b) | (math.ceil(self.a * 255) << 24)
end

Color.White = Color(0xffffff)
Color.Black = Color(0)
Color.Gray = Color(0x808080)
Color.LightGray = Color(0xd3d3d3)
Color.DarkGray = Color(0x303030)
Color.Red = Color(0xff0000)
Color.LightRed = Color(0xff8080)
Color.DarkRed = Color(0x290000)
Color.Yellow = Color(0xf2ab26)
Color.LightYellow = Color(0xffff7f)
Color.DarkYellow = Color(0x332b00)
Color.Green = Color(0x00ff00)
Color.LightGreen = Color(0x7fff7f)
Color.LighterGreen = Color(0xbfff7f)
Color.LightestGreen = Color(0xdfff7f)
Color.DarkGreen = Color(0x007f00)
Color.Cyan = Color(0x00ffff)
Color.LightCyan = Color(0x7fffff)
Color.DarkCyan = Color(0x007f7f)
Color.Magenta = Color(0xff00ff)
Color.LightMagenta = Color(0xff7fff)
Color.DarkMagenta = Color(0x7f007f)
Color.Blue = Color(0x0000ff)
Color.LightBlue = Color(0x7f7fff)
Color.DarkBlue = Color(0x00007f)
Color.Turquoise = Color(0x40e0d0)
Color.LightTurquoise = Color(0x80ffff)
Color.DarkTurquoise = Color(0x008080)
Color.SkyBlue = Color(0x87bedb)
Color.Gold = Color(0xffd700)
Color.DarkGold = Color(0x8b7500)
Color.GUI = {}
Color.GUI.window_features = Color(0xff333333)
Color.GUI.window_background = Color(0xff1e1e1e)
Color.GUI.window_sidebar = Color(0xff252526)
Color.GUI.window_hl = Color(0xff555555)
Color.GUI.text = Color(0xff858585)
Color.GUI.text_hl = Color(0xffc6c6c6)
Color.GUI.comment = Color(0xff6a9955)
Color.GUI.define = Color(0xffc586c0)
Color.GUI.keyword = Color(0xff569cd6)
Color.GUI.type = Color(0xff4ec9b0)
Color.GUI.fn = Color(0xffdcdcaa)
Color.GUI.close_button = Color(0xff7f0a1a)
Color.GUI.outline = Color(0xff3a3a3a)

return Color